from django.contrib import auth
from django.contrib import messages
from django.core.urlresolvers import reverse
from django import http
from django import shortcuts


def home(request):
    return shortcuts.render(request, 'devsite/home.html', {
        'user': request.user,
    })

def logout(request):
    auth.logout(request)
    messages.success(request, u"Logout successful.")
    return http.HttpResponseRedirect(reverse('home'))



# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois

from __future__ import absolute_import, unicode_literals


from django.conf import settings

import appconf


class AuthGroupeXConf(appconf.AppConf):
    """Global, django-appconf based config for authgroupex."""

    # Pre-shared key for authenticating our site to remote
    KEY = ''

    # Url of the endpoint
    ENDPOINT = 'https://www.polytechnique.org/auth-groupex/utf8'

    # Return URL where the AuthGroupeX server should user back
    RETURN_URL = ''
    def configure_return_url(self, value):
        return value or settings.LOGIN_URL

    # Fields to fetch from remote; order matters.
    FIELDS = ('username', 'firstname', 'lastname', 'email')

    # The AuthGroupeX permissions that enable the 'is_admin' flag on this site
    SUPERADMIN_PERMS = ()

    # The AuthGroupeX permissions that enable the 'is_staff' flag on this site
    STAFF_PERMS = ()

    # Whether accounts for now-dead users should switch to is_active=False
    DISABLE_DEADS = False

    # Name of the group to query as
    GROUP = ''

    # Maps a auth-groupex perm to a local group name
    MAP_GROUPS = {}

    # Model to store users to
    USER_MODEL = 'auth.User'

    # Group to store groups to
    GROUP_MODEL = 'auth.Group'

    # Whether to enable debug
    FAKE = False
    def configure_fake(self, value):
        return value or settings.DEBUG

    class Meta:
        prefix = 'authgroupex'
        # Forward setting lookups to base django.conf.settings
        proxy = True



# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois


"""Provide a AuthGroupeX authentication backend for Django."""

from __future__ import absolute_import


from .auth import AuthGroupeXBackend
from .auth import PERM_ADMIN, PERM_GROUP_ADMIN, PERM_GROUP_MEMBER, PERM_USER
from .exceptions import AuthGroupeXError, InvalidAuth
from .views import AuthGroupeXBaseView, AuthGroupeXUniqueView

__author__ = u"Raphaël Barrois <raphael.barrois+djauthgroupex@polytechnique.org>"
__version__ = '0.3.1'



# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois


from __future__ import absolute_import, unicode_literals

import logging

from django.core import exceptions
from django.db import models


from . import conf


PERM_USER = 'user'
PERM_GROUP_MEMBER = 'grpmember'
PERM_GROUP_ADMIN = 'grpadmin'
PERM_ADMIN = 'admin'

PERM_LEVELS = (
    PERM_USER,
    PERM_GROUP_MEMBER,
    PERM_GROUP_ADMIN,
    PERM_ADMIN,
)


logger = logging.getLogger(__name__)


def get_model(model_name):
    """Retrieve a django model.

    This handles:
    - Explicit models.Model subclass
    - Absolute dotted path to import the model
    """
    if isinstance(model_name, type) and issubclass(models.Model, model_name):
        return model_name

    # Not a Model or a Model instance, must be a class path
    if '.' not in model_name:
        raise ValueError("Invalid model name %s: should include module name."
                % model_name)

    app, cls = model_name.rsplit('.', 1)
    return models.get_model(app, cls)


class AuthResult(object):
    """Result of an authentication query."""

    def __init__(self, success, data=None):
        if not data:
            data = {}
        self.success = success
        self.data = data
        self.perms = self._setup_perms(self.data)

    def __repr__(self):
        return '<AuthResult: %s [%s / %s]>' % (
            'OK' if self.success else 'KO',
            self.data,
            self.perms,
        )

    def _setup_perms(self, data):
        perms = set()
        perms.add(PERM_USER)

        if 'perms' in data:
            perms.add(data['perms'])

        if 'grpauth' in data:
            if data['grpauth'] == 'admin':
                perms.add(PERM_GROUP_ADMIN)
                perms.add(PERM_GROUP_MEMBER)
            elif self._data['grpauth'] == 'membre':
                perms.add(PERM_GROUP_MEMBER)
        return perms

    @property
    def username(self):
        return self.data.get('username', '')

    @property
    def firstname(self):
        return self.data.get('firstname', '')

    @property
    def lastname(self):
        return self.data.get('lastname', '')

    @property
    def promo(self):
        return self.data.get('promo', '')

    @property
    def email(self):
        return self.data.get('email', '')

    @property
    def is_dead(self):
        return bool(self.data.get('deathdate', ''))

    @property
    def is_admin(self):
        return PERM_ADMIN in self.perms

    def has_perm(self, perm):
        return perm in self.perms


class AuthGroupeXMixin(object):
    def __init__(self, config=None, *args, **kwargs):
        super(AuthGroupeXMixin, self).__init__(*args, **kwargs)
        self.config = config or conf.AuthGroupeXConf()

    # Public API
    # ==========

    def authenticate(self, **kwargs):
        """Create a user if the authgroupex data has been passed.

        This data should be present in the 'authgroupex' keyword argument.
        """
        if 'authgroupex' in kwargs:
            auth_data = kwargs['authgroupex']
        else:
            logger.info('Trying to authenticate, no authgroupex in data.')
            return None

        if not auth_data.username:
            logger.error('Received a AuthResult object without a username.')
            return None

        try:
            user = self._fetch_user(auth_data.username)
        except exceptions.ObjectDoesNotExist:
            try:
                user = self._create_user_from_auth_data(auth_data)
            except ValueError:
                logger.warning('Received authgroupex with invalid name %s',
                        auth_data.username)
                return None

        self._update_user(user, auth_data)
        return user

    # Required extension points
    # =========================

    def get_user(self, user_id):
        raise NotImplementedError()

    def _fetch_user(self, username):
        raise NotImplementedError()

    def _create_user(self, username):
        raise NotImplementedError()

    # Optional extension points
    # =========================

    def _set_staff(self, user, is_staff):
        if hasattr(user, 'is_staff'):
            user.is_staff = is_staff

    def _set_superuser(self, user, is_superuser):
        if hasattr(user, 'is_superuser'):
            user.is_superuser = is_superuser

    def _set_active(self, user, is_active):
        if hasattr(user, 'is_active'):
            user.is_active = is_active

    def _update_profile(self, user, auth_data):
        """Update fields of the profile according to auth-groupe-x data."""
        pass

    def _update_groups(self, user, auth_data):
        pass

    # Internals
    # =========

    def _update_perms(self, user, auth_data):
        # Handle staff status
        if self.config.STAFF_PERMS:
            self._set_staff(user, any(
                auth_data.has_perm(perm) for perm in self.config.STAFF_PERMS))

        # Handle superadmins
        if self.config.SUPERADMIN_PERMS:
            is_superuser = any(
                auth_data.has_perm(perm) for perm in self.config.SUPERADMIN_PERMS)

            self._set_superuser(user, is_superuser)
            if is_superuser:
                self._set_staff(user, True)

        # Handle active status
        if auth_data.is_dead and self.config.DISABLE_DEADS:
            self._set_active(user, False)

    def _update_user(self, user, auth_data):
        """Update various fields of the user according to auth-groupe-x data."""
        self._update_profile(user, auth_data)
        self._update_perms(user, auth_data)
        self._update_groups(user, auth_data)

        # Refresh DB user
        user.save()
        logger.info('Updated user %s', user.get_username())

    def _create_user_from_auth_data(self, auth_data):
        """Create a new Django user from AuthGroupeX data.

        This only sets the basic username field;
        groups and other data are handled by the update_user method.
        """
        username = auth_data.username
        user = self._create_user(username)
        user.set_unusable_password()
        logger.info('Created a new user with username %s', username)
        return user


class AuthGroupeXBackend(AuthGroupeXMixin):
    """Authentication backend for auth-groupe-x"""

    supports_anonymous_user = False
    supports_object_permissions = False

    def __init__(self, config=None, *args, **kwargs):
        super(AuthGroupeXBackend, self).__init__(*args, **kwargs)
        self.user_model = get_model(self.config.USER_MODEL)

    def get_user(self, user_id):
        """Retrieve a user by ID.

        Args:
            user_id: int, the ID of the user

        Returns:
            Either an instance of self.config.USER_MODEL or None
        """
        try:
            return self.user_model.objects.get(pk=user_id)
        except self.user_model.DoesNotExist:
            return None

    def _fetch_user(self, username):
        return self.user_model.objects.get(username=username)

    def _create_user(self, username):
        return self.user_model.objects.create(username=username, is_active=True)

    def _update_profile(self, user, auth_data):
        """Update fields of the profile according to auth-groupe-x data."""
        # Update basic profile data
        if auth_data.firstname:
            user.first_name = auth_data.firstname
        if auth_data.lastname:
            user.last_name = auth_data.lastname
        if auth_data.email:
            user.email = auth_data.email

        if getattr(self.config, 'PROFILE_CLASS', ''):
            profile_model = get_model(self.config.PROFILE_CLASS)
            try:
                profile = user.get_profile()
            except profile_model.DoesNotExist:
                profile = profile_model.objects.create(user=user)
            if auth_data.promo:
                profile.promo = auth_data.promo
            profile.save()

    def _update_groups(self, user, auth_data):
        if not self.config.MAP_GROUPS:
            return

        group_model = get_model(self.config.GROUP_MODEL)
        new_group_names = set()
        old_group_names = set()
        for perm in PERM_LEVELS:
            if auth_data.has_perm(perm):
                new_group_names |= set(self.config.MAP_GROUPS.get(perm, []))
            else:
                old_group_names |= set(self.config.MAP_GROUPS.get(perm, []))

        new_groups = list(group_model.objects.filter(name__in=new_group_names))
        old_groups = list(group_model.objects.filter(name__in=old_group_names))

        if old_groups:
            logger.info(u"Removing user %s from groups %s", user, new_groups)
            user.groups.remove(*list(old_groups))

        if new_groups:
            logger.info(u"Adding user %s to groups %s", user, new_groups)
            user.groups.add(*list(new_groups))



# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois

from __future__ import absolute_import, unicode_literals


import hashlib
import random

from django import http
from django.contrib import auth as django_auth
from django.core.urlresolvers import reverse

from . import auth
from . import conf
from . import exceptions


class URLFormatter(object):
    """Handles URL formatting and parsing."""

    def __init__(self, config=None, *args, **kwargs):
        super(URLFormatter, self).__init__(*args, **kwargs)
        self.config = config or conf.AuthGroupeXConf()

    def make_url(self, request, return_url=None):
        """Compute the remote URL.
        
        Args:
            request: http.HttpRequest, the request to use for session storage
                and current URL retrieving
            return_url: str, the optional return URL to use; if empty, the
                current URL of the request will be used.

        Returns:
            the new url.
        """
        challenge = hashlib.sha1(
                b''.join(chr(random.randrange(0, 256)) for i in range(64)))
        challenge = challenge.hexdigest()

        request.session['authgroupex-challenge'] = challenge

        sig = hashlib.md5(challenge + self.config.KEY).hexdigest()
        query = http.QueryDict('', mutable=True)
        query['challenge'] = challenge
        query['pass'] = sig
        query['url'] = request.build_absolute_uri(return_url)

        endpoint = self.config.ENDPOINT
        if self.config.FAKE:
            endpoint = reverse(endpoint)

        url = endpoint + '?' + query.urlencode()
        return url

    def parse_return(self, request):
        """Parse the returned data.

        Args:
            request: the request received on the return URL

        Returns:
            An AuthResult object.
        """
        if 'auth' not in request.GET:
            raise exceptions.InvalidAuth("request.GET lacks a 'auth' field.")

        if 'authgroupex-challenge' not in request.session:
            raise exceptions.InvalidAuth(
                    "request.session lacks a 'authgroupex-challenge' field.")

        data = {}
        check_str = '1{challenge}{key}'.format(
                challenge=request.session['authgroupex-challenge'],
                key=self.config.KEY)

        for field in self.config.FIELDS:
            if field not in request.GET:
                return auth.AuthResult(success=False)
            data[field] = request.GET[field]
            check_str += data[field]

        check_str += '1'
        check_str = check_str.encode('utf-8')
        if hashlib.md5(check_str).hexdigest() != request.GET['auth']:
            raise exceptions.InvalidAuth(
                    "Invalid signature in authgroupex response.")

        return auth.AuthResult(success=True, data=data)


# Expected behaviour:
# 1) User goes to /foobar/
# 2) The view for /foobar/ has a @login_required decorator
# 3) @login_required redirects to <LOGIN_URL>?next=/foobar/
# 4) (optionnally) the user clicks on a link that brings him to AuthGroupeXBaseView.login_begin_view with ?next=/foobar/
# 5) AuthGroupeXBaseView.login_begin_view redirects to https://x.org/...&url=AuthGroupeXBaseView.login_return_view?next=/foobar/
# 6) X.org authenticates and redirects to AuthGroupeXBaseView.login_return_view?user=..&next=/foobar/
# 7) AuthGroupeXBaseView.login_return_view sends to /foobar/


class AuthGroupeXBaseView(object):
    """Base view class to use for login.

    Provides separated 'login' and 'return' views.
    """

    def __init__(self, url_formatter=None, config=None, *args, **kwargs):
        super(AuthGroupeXBaseView, self).__init__(*args, **kwargs)
        self.url_formatter = url_formatter or URLFormatter()
        self.config = config or conf.AuthGroupeXConf()
        self.include_return_param = False

    # We need to redefine builtin 'next'
    # pylint: disable=W0622
    def build_return_url(self, request, next=None):
        """Builds the return URL for a request."""
        url = self.config.RETURN_URL
        query = http.QueryDict('', mutable=True)
        if next:
            query['next'] = next

        if self.include_return_param:
            query['auth_groupex_return'] = '1'

        url += '?' + query.urlencode()
        return self.url_formatter.make_url(request, url)

    def login_begin_view(self, request, next=None):
        """View initiating the authgroupex authentication."""
        next = next or request.GET.get('next')
        auth_url = self.build_return_url(request, next)
        return http.HttpResponseRedirect(auth_url)

    def admin_login(self, request, extra_context=None):
        """View to use instead of the standard admin login page."""
        return self.login_begin_view(request, next=request.get_full_path())
    # pylint: enable=W0622

    def login_return_view(self, request):
        """View handling the 'return' data from authgroupex."""
        data = self.url_formatter.parse_return(request)
        user = django_auth.authenticate(authgroupex=data)
        if user and user.is_active:
            django_auth.login(request, user)
            if 'next' in request.GET:
                redirect_to = request.GET['next']
            else:
                redirect_to = self.config.LOGIN_REDIRECT_URL

            return http.HttpResponseRedirect(redirect_to)
        else:
            return http.HttpResponse('Internal error.', status=500)


class AuthGroupeXUniqueView(AuthGroupeXBaseView):
    """View class providing a unified login/return view."""

    def __init__(self, *args, **kwargs):
        super(AuthGroupeXUniqueView, self).__init__(*args, **kwargs)
        self.include_return_param = True

    # We need to redefine builtin 'next'
    # pylint: disable=W0622
    def login_view(self, request, next=None):
        """View dispatching the request to the adequate processing view."""
        if request.GET.get('auth_groupex_return') == '1':
            return self.login_return_view(request)
        else:
            return self.login_begin_view(request, next)
    # pylint: enable=W0622



from django.conf.urls import patterns, url

from . import conf
from . import views


urlpatterns = patterns('',
    url(r'^$', views.AuthGroupeXUniqueView().login_view, name='login'),
)


config = conf.AuthGroupeXConf()

if config.FAKE:
    urlpatterns += patterns('django_authgroupex.fake.views',
        url(r'^fake/validate/$', 'endpoint', name='fake_endpoint'),
        url(r'^fake/fill/$', 'login_view', name='fake_fill'),
    )



# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois

from __future__ import absolute_import, unicode_literals


class AuthGroupeXError(Exception):
    """Base class for AuthGroupeX errors."""


class ProcessError(AuthGroupeXError):
    """Raised when an error occured in the request-response process."""


class InvalidAuth(ProcessError):
    """Raised when the 'auth' field of a response was invalid."""


class UsernameViolation(AuthGroupeXError):
    """Raised when an invalid username was submitted."""




# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois

from __future__ import absolute_import, unicode_literals


"""Provide fake login views."""


from django.contrib import messages
from django.core.urlresolvers import resolve, reverse
from django.shortcuts import redirect, render

from .. import conf
from . import forms


config = conf.AuthGroupeXConf()


def smart_reverse(request, name):
    """Workaround until Django 1.5."""
    match = resolve(request.path)
    if match.namespace:
        prefix = '%s:' % match.namespace
    else:
        prefix = ''

    return reverse('%s%s' % (prefix, name), current_app=match.app_name)


def endpoint(request):
    """The view receiving the incoming request."""
    form = forms.EndPointForm(data=request.GET or None)

    if form.is_valid():
        next_url = smart_reverse(request, 'fake_fill') + '?' + form.build_query()
        messages.success(request, u"The auth-groupex request was properly signed.")
        return redirect(next_url)

    else:
        messages.error(request, u"The auth-groupex request couldn't be verified.")
        return render(request, 'authgroupex/fake_endpoint_fail.html', {'form': form})


def login_view(request):
    form = forms.LoginForm(data=request.POST or None, fields=config.FIELDS)

    if form.is_valid():
        reply_querystring = form.build_reply(request.GET['challenge'])
        reply_url = request.GET['url']
        separator = '&' if '?' in reply_url else '?'
        messages.success(request, u"Redirecting to %s with a signed request." % reply_url)
        return redirect(reply_url + separator + reply_querystring)
    else:
        return render(request, 'authgroupex/fake_fill.html', {'form': form})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois

import os
import re
import sys

from setuptools import setup
from distutils import cmd

root = os.path.abspath(os.path.dirname(__file__))

def get_version(*module_dir_components):
    version_re = re.compile(r"^__version__ = ['\"](.*)['\"]$")
    module_root = os.path.join(root, *module_dir_components)
    module_init = os.path.join(module_root, '__init__.py')
    with open(module_init, 'r') as f:
        for line in f:
            match = version_re.match(line[:-1])
            if match:
                return match.groups()[0]
    return '0.1.0'


PACKAGE = 'django_authgroupex'
VERSION = get_version(PACKAGE)


setup(
    name="django_authgroupex",
    version=VERSION,
    description="An authentication backend for Django based on Polytechnique.org's auth-groupe-x SSO protocol.",
    author="Raphaël Barrois",
    author_email="raphael.barrois+djauthgroupex@polytechnique.org",
    url='http://github.com/rbarrois/django-authgroupex',
    download_url='http://pypi.python.org/pypi/django-authgroupex/',
    keywords=['sso', 'authentication', 'django', 'authgroupex'],
    packages=['django_authgroupex', 'django_authgroupex.fake'],
    package_data={
        'django_authgroupex': [
            os.path.join('templates', 'authgroupex', '*.html'),
            os.path.join('static', 'authgroupex', '*.js'),
            os.path.join('static', 'authgroupex', '*.css'),
        ],
    },
    license='BSD',
    setup_requires=[
        'setuptools>=0.8',
    ],
    install_requires=[
        'Django>=1.3',
        'django_appconf',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
)




# -*- coding: utf-8 -*-
# This code is distributed under the two-clause BSD license.
# Copyright (c) 2012-2013 Raphaël Barrois

from __future__ import absolute_import, unicode_literals

import hashlib
import urllib

from django import forms
from django import http

from .. import conf


PERM_CHOICES = (
    ('admin', "Admin"),
    ('user', "User"),
)

GROUP_AUTH_CHOICES = (
    ('grpmember', "Group member"),
    ('grpadmin', "Group admin"),
)


DEFAULT_FIELDS = {
    'username': forms.SlugField(label="Username", required=False),
    'firstname': forms.CharField(label="First name", required=False),
    'lastname': forms.CharField(label="Last name", required=False),
    'perms': forms.ChoiceField(choices=PERM_CHOICES, label="Permissions", required=False),
    'grpauth': forms.ChoiceField(choices=GROUP_AUTH_CHOICES, label="Group access level", required=False),
}


class LoginForm(forms.Form):
    def __init__(self, fields=(), *args, **kwargs):
        super(LoginForm, self).__init__(*args, **kwargs)

        for name in fields:
            try:
                form_field = DEFAULT_FIELDS[name]
            except KeyError:
                form_field = forms.CharField(label=name.capitalize(), required=False)
            self.fields[name] = form_field

    def build_reply(self, challenge):
        config = conf.AuthGroupeXConf()

        auth_parts = ['1', challenge, config.KEY]
        reply = http.QueryDict('', mutable=True)

        for field in config.FIELDS:
            value = self.cleaned_data[field]
            reply[field] = value
            auth_parts.append(value)

        auth_parts.append('1')
        auth = ''.join(auth_parts).encode('utf8')

        reply['auth'] = hashlib.md5(auth).hexdigest()

        return reply.urlencode()


long_text = forms.TextInput(attrs={'size': 60})


class EndPointForm(forms.Form):
    url = forms.CharField(max_length=600, widget=long_text, help_text="The return URL")
    challenge = forms.CharField(max_length=64, widget=long_text,
            help_text="A random challenge, crafted specifically for this request")
    _pass = forms.CharField(max_length=64, widget=long_text,
            help_text="Signature : md5(challenge + private_key)")

    def __init__(self, *args, **kwargs):
        super(EndPointForm, self).__init__(*args, **kwargs)
        self.fields['pass'] = self.fields.pop('_pass')

    def clean_url(self):
        return urllib.unquote(self.cleaned_data['url'])

    def clean(self):
        config = conf.AuthGroupeXConf()
        challenge = self.cleaned_data['challenge']
        password = self.cleaned_data['pass']

        expected_pass = hashlib.md5(challenge + config.KEY).hexdigest()

        if expected_pass != password:
            raise forms.ValidationError("Expected signature %s, got %s" % (expected_pass, password))
        return self.cleaned_data

    def build_query(self):
        """Build the query part for the 'next' URL."""
        url = self.cleaned_data['url']
        challenge = self.cleaned_data['challenge']

        query = http.QueryDict('', mutable=True)
        query['challenge'] = challenge
        query['url'] = url

        return query.urlencode()



from django.conf.urls import patterns, include, url

from django.contrib import admin
admin.autodiscover()
# Use authgroupex login on admin site
admin.site.login_template = 'authgroupex/admin_login.html'

urlpatterns = patterns('',
    url(r'^$', 'dev.devsite.views.home', name='home'),
    url(r'^xorgauth/', include('django_authgroupex.urls', namespace='authgroupex')),

    url(r'^admin/', include(admin.site.urls)),
    url(r'^logout/', 'dev.devsite.views.logout', name='logout'),
)
